import busboy from 'busboy';
import createError from 'http-errors';
import objectPath from 'object-path';
import GRAPHQL_MULTIPART_REQUEST_SPEC_URL from './GRAPHQL_MULTIPART_REQUEST_SPEC_URL.mjs';
import ignoreStream from './ignoreStream.mjs';
import Upload from './Upload.mjs';
import { WriteStream } from 'fs-capacitor';
export default async function processRequest(request, response, { maxFieldSize = 1000000, maxFileSize = Infinity, maxFiles = Infinity, } = {}) {
    return new Promise((resolve, reject) => {
        let released;
        let exitError;
        let operations;
        let operationsPath;
        let map;
        const parser = busboy({
            headers: request.headers,
            defParamCharset: 'utf8',
            limits: {
                fieldSize: maxFieldSize,
                fields: 2,
                fileSize: maxFileSize,
                files: maxFiles,
            },
        });
        function exit(error, isParserError = false) {
            if (exitError)
                return;
            exitError = error;
            if (map)
                for (const upload of map.values())
                    if (!upload.file)
                        upload.reject(exitError);
            isParserError ? parser.destroy() : parser.destroy(exitError);
            request.unpipe(parser);
            setImmediate(() => {
                request.resume();
            });
            reject(exitError);
        }
        parser.on('field', (fieldName, value, { valueTruncated }) => {
            if (valueTruncated)
                return exit(createError(413, `The ‘${fieldName}’ multipart field value exceeds the ${maxFieldSize} byte size limit.`));
            switch (fieldName) {
                case 'operations':
                    try {
                        operations = JSON.parse(value);
                    }
                    catch (error) {
                        return exit(createError(400, `Invalid JSON in the ‘operations’ multipart field (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                    }
                    if (typeof operations !== 'object' || !operations)
                        return exit(createError(400, `Invalid type for the ‘operations’ multipart field (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                    operationsPath = objectPath(operations);
                    break;
                case 'map': {
                    if (!operations)
                        return exit(createError(400, `Misordered multipart fields; ‘map’ should follow ‘operations’ (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                    let parsedMap;
                    try {
                        parsedMap = JSON.parse(value);
                    }
                    catch (error) {
                        return exit(createError(400, `Invalid JSON in the ‘map’ multipart field (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                    }
                    if (typeof parsedMap !== 'object' ||
                        !parsedMap ||
                        Array.isArray(parsedMap))
                        return exit(createError(400, `Invalid type for the ‘map’ multipart field (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                    const mapEntries = Object.entries(parsedMap);
                    if (mapEntries.length > maxFiles)
                        return exit(createError(413, `${maxFiles} max file uploads exceeded.`));
                    map = new Map();
                    for (const [fieldName, paths] of mapEntries) {
                        if (!Array.isArray(paths))
                            return exit(createError(400, `Invalid type for the ‘map’ multipart field entry key ‘${fieldName}’ array (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                        map.set(fieldName, new Upload());
                        for (const [index, path] of paths.entries()) {
                            if (typeof path !== 'string')
                                return exit(createError(400, `Invalid type for the ‘map’ multipart field entry key ‘${fieldName}’ array index ‘${index}’ value (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                            try {
                                operationsPath.set(path, map.get(fieldName));
                            }
                            catch (error) {
                                return exit(createError(400, `Invalid object path for the ‘map’ multipart field entry key ‘${fieldName}’ array index ‘${index}’ value ‘${path}’ (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
                            }
                        }
                    }
                    resolve(operations);
                }
            }
        });
        parser.on('file', (fieldName, stream, { filename, encoding, mimeType: mimetype }) => {
            if (!map) {
                ignoreStream(stream);
                return exit(createError(400, `Misordered multipart fields; files should follow ‘map’ (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
            }
            const upload = map.get(fieldName);
            if (!upload) {
                ignoreStream(stream);
                return;
            }
            let fileError;
            const capacitor = new WriteStream();
            capacitor.on('error', () => {
                stream.unpipe();
                stream.resume();
            });
            stream.on('limit', () => {
                fileError = createError(413, `File truncated as it exceeds the ${maxFileSize} byte size limit.`);
                stream.unpipe();
                capacitor.destroy(fileError);
            });
            stream.on('error', (error) => {
                fileError = error;
                stream.unpipe();
                capacitor.destroy(fileError);
            });
            const file = {
                filename,
                mimetype,
                encoding,
                createReadStream(options) {
                    const error = fileError || (released ? exitError : null);
                    if (error)
                        throw error;
                    return capacitor.createReadStream(options);
                },
                capacitor,
            };
            Object.defineProperty(file, 'capacitor', {
                enumerable: false,
                configurable: false,
                writable: false,
            });
            stream.pipe(capacitor);
            upload.resolve(file);
        });
        parser.once('filesLimit', () => exit(createError(413, `${maxFiles} max file uploads exceeded.`)));
        parser.once('finish', () => {
            request.unpipe(parser);
            request.resume();
            if (!operations)
                return exit(createError(400, `Missing multipart field ‘operations’ (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
            if (!map)
                return exit(createError(400, `Missing multipart field ‘map’ (${GRAPHQL_MULTIPART_REQUEST_SPEC_URL}).`));
            for (const upload of map.values())
                if (!upload.file)
                    upload.reject(createError(400, 'File missing in the request.'));
        });
        parser.on('error', (error) => {
            return exit(error, true);
        });
        response.once('close', () => {
            released = true;
            if (map)
                for (const upload of map.values())
                    if (upload.file)
                        upload.file.capacitor.release();
        });
        request.once('close', () => {
            if (!request.readableEnded)
                exit(createError(499, 'Request disconnected during file upload stream parsing.'));
        });
        request.pipe(parser);
    });
}
